""" pyintesishome Module """
from .exceptions import IHAuthenticationError, IHConnectionError
from .intesisbase import IntesisBase
from .intesisbox import IntesisBox
from .intesishome import IntesisHome
from .intesishomelocal import IntesisHomeLocal



"""Base class for Intesis controllers."""
import asyncio
import logging
from asyncio.exceptions import IncompleteReadError
from asyncio.streams import StreamReader, StreamWriter

import aiohttp

from .const import (
    COMMAND_MAP,
    CONFIG_MODE_BITS,
    DEVICE_INTESISHOME,
    ERROR_MAP,
    INTESIS_MAP,
    INTESIS_NULL,
    OPERATING_MODE_BITS,
)
from .helpers import twos_complement_16bit, uint32

_LOGGER = logging.getLogger("pyintesishome")


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-public-methods
class IntesisBase:
    """Base class for Intesis controllers."""

    def __init__(
        self,
        username=None,
        password=None,
        host=None,
        loop=None,
        websession=None,
        device_type=DEVICE_INTESISHOME,
    ):
        """Initialize IntesisBox controller."""
        # Select correct API for device type
        self._username = username
        self._password = password
        self._host = host
        self._device_type = device_type
        self._devices = {}
        self._connected = False
        self._connecting = False
        self._connection_retries = 0
        self._update_callbacks = []
        self._keepalive_task: asyncio.Task = None
        self._receive_task: asyncio.Task = None
        self._error_message = None
        self._web_session = websession
        self._own_session = False
        self._controller_id = username
        self._controller_name = username
        self._writer: StreamWriter = None
        self._reader: StreamReader = None
        self._received_response: asyncio.Event = asyncio.Event()
        self._data_delimiter = b"}}"

        if loop:
            _LOGGER.debug("Using the provided event loop")
            self._event_loop = loop
        else:
            _LOGGER.debug("Getting the running loop from asyncio")
            self._event_loop = asyncio.get_running_loop()

        if not self._web_session:
            _LOGGER.debug("Creating new websession")
            self._web_session = aiohttp.ClientSession()
            self._own_session = True

    async def _set_value(self, device_id, uid, value):
        """Internal method to send a value to the device."""
        raise NotImplementedError()

    async def _send_command(self, command: str):
        try:
            _LOGGER.debug("Sending command %s", command)
            self._received_response.clear()
            if self._writer:
                self._writer.write(command.encode("ascii"))
                await self._writer.drain()
                try:
                    await asyncio.wait_for(
                        self._received_response.wait(),
                        timeout=5.0,
                    )
                except asyncio.TimeoutError:
                    print("oops took longer than 5s!")
        except OSError as exc:
            _LOGGER.error("%s Exception. %s / %s", type(exc), exc.args, exc)

    async def _data_received(self):
        try:
            while self._reader:
                raw_data = await self._reader.readuntil(self._data_delimiter)
                if not raw_data:
                    break
                data = raw_data.decode("ascii")
                _LOGGER.debug("Received: %s", data)
                await self._parse_response(data)

                if not self._received_response.is_set():
                    _LOGGER.debug("Resolving set_value's await")
                    self._received_response.set()
        except IncompleteReadError:
            _LOGGER.info(
                "pyIntesisHome lost connection to the %s server.", self._device_type
            )
        except asyncio.CancelledError:
            pass
        except (
            TimeoutError,
            ConnectionResetError,
            OSError,
        ) as exc:
            _LOGGER.error(
                "pyIntesisHome lost connection to the %s server. Exception: %s",
                self._device_type,
                exc,
            )
        finally:
            self._connected = False
            self._connecting = False
            await self._send_update_callback()

    def _update_device_state(self, device_id, uid, value):
        """Internal method to update the state table of IntesisHome/Airconwithme devices."""
        device_id = str(device_id)

        if uid in INTESIS_MAP:
            # If the value is null (32768), set as None
            if value == INTESIS_NULL:
                self._devices[device_id][INTESIS_MAP[uid]["name"]] = None
            else:
                # Translate known UIDs to configuration item names
                value_map = INTESIS_MAP[uid].get("values")
                if value_map:
                    self._devices[device_id][INTESIS_MAP[uid]["name"]] = value_map.get(
                        value, value
                    )
                else:
                    self._devices[device_id][INTESIS_MAP[uid]["name"]] = value
        else:
            # Log unknown UIDs
            self._devices[device_id][f"unknown_uid_{uid}"] = value

    def _update_rssi(self, device_id, rssi):
        """Internal method to update the wireless signal strength."""
        if rssi and str(device_id) in self._devices:
            self._devices[str(device_id)]["rssi"] = rssi

    async def connect(self):
        """Public method for connecting to API"""
        raise NotImplementedError()

    async def stop(self):
        """Public method for shutting down connectivity."""
        self._connected = False
        await self._cancel_task_if_exists(self._receive_task)
        await self._cancel_task_if_exists(self._keepalive_task)
        if self._writer:
            self._writer.close()
            await self._writer.wait_closed()

        if self._own_session:
            await self._web_session.close()

    @staticmethod
    async def _cancel_task_if_exists(task: asyncio.Task):
        if task:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def poll_status(self, sendcallback=False):
        """Public method to query IntesisHome for state of device.
        Notifies subscribers if sendCallback True."""
        raise NotImplementedError()

    def get_devices(self):
        """Public method to return the state of all IntesisHome devices"""
        return self._devices

    def get_device(self, device_id):
        """Public method to return the state of the specified device"""
        return self._devices.get(str(device_id))

    def get_device_property(self, device_id, property_name):
        """Public method to get a property of the specified device"""
        return self._devices[str(device_id)].get(property_name)

    def get_run_hours(self, device_id) -> str:
        """Public method returns the run hours of the IntesisHome controller."""
        run_hours = self.get_device_property(device_id, "working_hours")
        return run_hours

    async def set_mode(self, device_id, mode: str):
        """Internal method for setting the mode with a string value."""
        mode_control = "mode"
        if "mode" not in self._devices[str(device_id)]:
            mode_control = "operating_mode"

        if mode in COMMAND_MAP[mode_control]["values"]:
            await self._set_value(
                device_id,
                COMMAND_MAP[mode_control]["uid"],
                COMMAND_MAP[mode_control]["values"][mode],
            )

    async def set_preset_mode(self, device_id, preset: str):
        """Internal method for setting the mode with a string value."""
        if preset in COMMAND_MAP["climate_working_mode"]["values"]:
            await self._set_value(
                device_id,
                COMMAND_MAP["climate_working_mode"]["uid"],
                COMMAND_MAP["climate_working_mode"]["values"][preset],
            )

    async def set_temperature(self, device_id, setpoint):
        """Public method for setting the temperature"""
        set_temp = uint32(setpoint * 10)
        await self._set_value(device_id, COMMAND_MAP["setpoint"]["uid"], set_temp)

    async def set_fan_speed(self, device_id, fan: str):
        """Public method to set the fan speed"""
        fan_map = self._get_fan_map(device_id)
        map_fan_speed_to_int = {v: k for k, v in fan_map.items()}
        await self._set_value(
            device_id, COMMAND_MAP["fan_speed"]["uid"], map_fan_speed_to_int[fan]
        )

    async def set_vertical_vane(self, device_id, vane: str):
        """Public method to set the vertical vane"""
        await self._set_value(
            device_id, COMMAND_MAP["vvane"]["uid"], COMMAND_MAP["vvane"]["values"][vane]
        )

    async def set_horizontal_vane(self, device_id, vane: str):
        """Public method to set the horizontal vane"""
        await self._set_value(
            device_id, COMMAND_MAP["hvane"]["uid"], COMMAND_MAP["hvane"]["values"][vane]
        )

    async def set_mode_heat(self, device_id):
        """Public method to set device to heat asynchronously."""
        await self.set_mode(device_id, "heat")

    async def set_mode_cool(self, device_id):
        """Public method to set device to cool asynchronously."""
        await self.set_mode(device_id, "cool")

    async def set_mode_fan(self, device_id):
        """Public method to set device to fan asynchronously."""
        await self.set_mode(device_id, "fan")

    async def set_mode_auto(self, device_id):
        """Public method to set device to auto asynchronously."""
        await self.set_mode(device_id, "auto")

    async def set_mode_dry(self, device_id):
        """Public method to set device to dry asynchronously."""
        await self.set_mode(device_id, "dry")

    async def set_power_off(self, device_id):
        """Public method to turn off the device asynchronously."""
        await self._set_value(
            device_id,
            COMMAND_MAP["power"]["uid"],
            COMMAND_MAP["power"]["values"]["off"],
        )

    async def set_power_on(self, device_id):
        """Public method to turn on the device asynchronously."""
        await self._set_value(
            device_id, COMMAND_MAP["power"]["uid"], COMMAND_MAP["power"]["values"]["on"]
        )

    def get_mode(self, device_id) -> str:
        """Public method returns the current mode of operation."""
        if "mode" in self._devices[str(device_id)]:
            return self._devices[str(device_id)]["mode"]
        if "operating_mode" in self._devices[str(device_id)]:
            return self._devices[str(device_id)]["operating_mode"]
        return None

    def get_mode_list(self, device_id) -> list:
        """Public method to return the list of device modes."""
        mode_list = []

        # By default, use config_mode_map to determine the available modes
        mode_map = self.get_device_property(device_id, "config_mode_map")
        mode_bits = CONFIG_MODE_BITS

        if "config_operating_mode" in self._devices[str(device_id)]:
            # If config_operating_mode is supplied, use that
            mode_map = self.get_device_property(device_id, "config_operating_mode")
            mode_bits = OPERATING_MODE_BITS

        # Generate the mode list from the map
        for mode_bit in mode_bits:
            if mode_map & mode_bit:
                mode_list.append(mode_bits.get(mode_bit))

        return mode_list

    def get_fan_speed(self, device_id):
        """Public method returns the current fan speed."""
        fan_map = self._get_fan_map(device_id)

        if "fan_speed" in self._devices[str(device_id)] and isinstance(fan_map, dict):
            fan_speed = self.get_device_property(device_id, "fan_speed")
            return fan_map.get(fan_speed, fan_speed)
        return None

    def get_fan_speed_list(self, device_id):
        """Public method to return the list of possible fan speeds."""
        fan_map = self._get_fan_map(device_id)
        if isinstance(fan_map, dict):
            return list(fan_map.values())
        return None

    def get_device_name(self, device_id) -> str:
        """Public method to get the name of a device."""
        return self.get_device_property(device_id, "name")

    def get_power_state(self, device_id) -> str:
        """Public method returns the current power state."""
        return self.get_device_property(device_id, "power")

    def get_instant_power_consumption(self, device_id) -> int:
        """Public method returns the current power state."""
        instant_power = self.get_device_property(device_id, "instant_power_consumption")
        if instant_power:
            return int(instant_power)
        return None

    def get_total_power_consumption(self, device_id) -> int:
        """Public method returns the current power state."""
        accumulated_power = self.get_device_property(
            device_id, "accumulated_power_consumption"
        )
        if accumulated_power:
            return int(accumulated_power)
        return None

    def get_cool_power_consumption(self, device_id) -> int:
        """Public method returns the current power state."""
        aquarea_cool = self.get_device_property(device_id, "aquarea_cool_consumption")
        if aquarea_cool:
            return int(aquarea_cool)
        return None

    def get_heat_power_consumption(self, device_id) -> int:
        """Public method returns the current power state."""
        aquarea_heat = self.get_device_property(device_id, "aquarea_heat_consumption")
        if aquarea_heat:
            return int(aquarea_heat)
        return None

    def get_tank_power_consumption(self, device_id) -> int:
        """Public method returns the current power state."""
        aquarea_tank = self.get_device_property(device_id, "aquarea_tank_consumption")
        if aquarea_tank:
            return int(aquarea_tank)
        return None

    def get_preset_mode(self, device_id) -> str:
        """Public method to get the current set preset mode."""
        return self.get_device_property(device_id, "climate_working_mode")

    def is_on(self, device_id) -> bool:
        """Return true if the controlled device is turned on"""
        return self.get_device_property(device_id, "power") == "on"

    def has_vertical_swing(self, device_id) -> bool:
        """Public method to check if the device has vertical swing."""
        vane_config = self.get_device_property(device_id, "config_vertical_vanes")
        vane_list = self.get_device_property(device_id, "vvane_list")
        return isinstance(vane_list, list) | bool(vane_config and vane_config > 1024)

    def has_horizontal_swing(self, device_id) -> bool:
        """Public method to check if the device has horizontal swing."""
        vane_config = self.get_device_property(device_id, "config_horizontal_vanes")
        vane_list = self.get_device_property(device_id, "hvane_list")
        return isinstance(vane_list, list) | bool(vane_config and vane_config > 1024)

    def has_setpoint_control(self, device_id) -> bool:
        """Public method to check if the device has setpoint control."""
        return "setpoint" in self._devices[str(device_id)]

    def get_setpoint(self, device_id) -> float:
        """Public method returns the target temperature."""
        setpoint = self.get_device_property(device_id, "setpoint")
        if setpoint:
            setpoint = int(setpoint) / 10
        return setpoint

    def get_temperature(self, device_id) -> float:
        """Public method returns the current temperature."""
        temperature = self.get_device_property(device_id, "temperature")
        if temperature:
            temperature = twos_complement_16bit(int(temperature)) / 10
        return temperature

    def get_outdoor_temperature(self, device_id) -> float:
        """Public method returns the current temperature."""
        outdoor_temp = self.get_device_property(device_id, "outdoor_temp")
        if outdoor_temp:
            outdoor_temp = twos_complement_16bit(int(outdoor_temp)) / 10
        return outdoor_temp

    def get_max_setpoint(self, device_id) -> float:
        """Public method returns the current maximum target temperature."""
        temperature = self.get_device_property(device_id, "setpoint_max")
        if temperature:
            temperature = int(temperature) / 10
        return temperature

    def get_min_setpoint(self, device_id) -> float:
        """Public method returns the current minimum target temperature."""
        temperature = self.get_device_property(device_id, "setpoint_min")
        if temperature:
            temperature = int(temperature) / 10
        return temperature

    def get_rssi(self, device_id) -> str:
        """Public method returns the current wireless signal strength."""
        rssi = self.get_device_property(device_id, "rssi")
        return rssi

    def get_vertical_swing(self, device_id) -> str:
        """Public method returns the current vertical vane setting."""
        swing = self.get_device_property(device_id, "vvane")
        return swing

    def get_horizontal_swing(self, device_id) -> str:
        """Public method returns the current horizontal vane setting."""
        swing = self.get_device_property(device_id, "hvane")
        return swing

    def get_error(self, device_id) -> str:
        """Public method returns the current error code + description."""
        error_code = self.get_device_property(device_id, "error_code")
        remote_code = ERROR_MAP[error_code]["code"]
        error_desc = ERROR_MAP[error_code]["desc"]
        return f"{remote_code}: {error_desc}"

    def _get_gen_value(self, device_id, name) -> str:
        """Internal method for getting generic value"""
        value = None
        if name in self._devices[str(device_id)]:
            value = self.get_device_property(device_id, name)
            _LOGGER.debug("%s = %s", name, value)
        else:
            _LOGGER.debug("No value for %s %s", device_id, name)
        return value

    def _get_gen_num_value(self, device_id, name):
        """Internal method for getting generic value and dividing by 10 if numeric"""
        value = self._get_gen_value(device_id, name)
        if isinstance(value, (int, float)):
            temperature = float(value / 10)
            return temperature
        return value

    def _set_gen_mode(self, device_id, gen_type, mode):
        """Internal method for setting the generic mode (gen_type in
        {operating_mode, climate_working_mode, tank, etc.}) with a string value"""
        if mode in COMMAND_MAP[gen_type]["values"]:
            self._set_value(
                device_id,
                COMMAND_MAP[gen_type]["uid"],
                COMMAND_MAP[gen_type]["values"][mode],
            )

    def _set_thermo_shift(self, device_id, name, value):
        """Public method to set thermo shift temperature."""
        min_shift = int(COMMAND_MAP[name]["min"])
        max_shift = int(COMMAND_MAP[name]["max"])

        if min_shift <= value <= max_shift:
            unsigned_value = uint32(value * 10)  # unsigned int 16 bit
            self._set_value(device_id, COMMAND_MAP[name]["uid"], unsigned_value)
        else:
            raise ValueError(
                f"Value for {name} has to be in range [{min_shift}],{max_shift}]"
            )

    @property
    def is_connected(self) -> bool:
        """Returns true if the TCP connection is established."""
        return self._connected

    @property
    def connection_retries(self) -> int:
        """Returns the number of connection retries."""
        return self._connection_retries

    @property
    def error_message(self) -> str:
        """Returns the last error message, or None if there were no errors."""
        return self._error_message

    @property
    def device_type(self) -> str:
        """Returns the device type (IntesisHome or airconwithme)."""
        return self._device_type

    @property
    def controller_id(self) -> str:
        """Returns an account/device identifier - Serial, MAC or username."""
        if self._controller_id:
            return self._controller_id.lower()
        return None

    @property
    def name(self) -> str:
        """Returns an account/device identifier - Serial, MAC or username."""
        if self._controller_name:
            return self._controller_name
        return None

    @property
    def is_disconnected(self) -> bool:
        """Returns true when the TCP connection is disconnected and idle."""
        return not self._connected and not self._connecting

    async def _send_update_callback(self, device_id=None):
        """Internal method to notify all update callback subscribers."""
        if self._update_callbacks:
            for callback in self._update_callbacks:
                await callback(device_id=device_id)

    def add_update_callback(self, method):
        """Public method to add a callback subscriber."""
        self._update_callbacks.append(method)

    def remove_update_callback(self, method):
        """Public method to add a callback subscriber."""
        self._update_callbacks.remove(method)

    def _get_fan_map(self, device_id):
        """Private method to get the fan_map."""
        raise NotImplementedError()

    async def _parse_response(self, decoded_data):
        """Private method to parse the API response."""
        raise NotImplementedError()



""" Main submodule for pyintesishome """

import asyncio
import logging

from .const import (
    COMMAND_MAP,
    DEVICE_INTESISHOME_LOCAL,
    INTESIS_MAP,
    LOCAL_CMD_GET_AVAIL_DP,
    LOCAL_CMD_GET_DP_VALUE,
    LOCAL_CMD_GET_INFO,
    LOCAL_CMD_LOGIN,
    LOCAL_CMD_SET_DP_VALUE,
)
from .exceptions import IHAuthenticationError, IHConnectionError
from .intesisbase import IntesisBase

_LOGGER = logging.getLogger("pyintesishome")


# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-public-methods
class IntesisHomeLocal(IntesisBase):
    """pyintesishome local class"""

    def __init__(self, host, username, password, loop=None, websession=None):
        device_type = (DEVICE_INTESISHOME_LOCAL,)
        self._session_id: str = ""
        self._datapoints: dict = {}
        self._scan_interval = 5
        self._device_id: str = ""
        self._values: dict = {}
        self._info: dict = {}
        self._update_task = None
        super().__init__(
            host=host,
            username=username,
            password=password,
            loop=loop,
            websession=websession,
            device_type=device_type,
        )

    async def _request_values(self) -> dict:
        """Get all entity values."""
        response = await self._request(LOCAL_CMD_GET_DP_VALUE, uid="all")
        self._values = {dpval["uid"]: dpval["value"] for dpval in response["dpval"]}
        return self._values

    async def _run_updater(self):
        """Run a loop that updates the values every _scan_interval."""
        while True:
            values = await self._request_values()
            for uid, value in values.items():
                self._update_device_state(self._device_id, uid, value)

            await self._send_update_callback(self._device_id)
            await asyncio.sleep(self._scan_interval)

    async def _authenticate(self) -> bool:
        """Authenticate using username and password."""
        response = await self._request(
            LOCAL_CMD_LOGIN, username=self._username, password=self._password
        )
        self._session_id = response["id"].get("sessionID")

    async def _request(self, command: str, **kwargs) -> dict:
        """Make a request."""
        payload = {
            "command": command,
            "data": {"sessionID": self._session_id, **kwargs},
        }

        async with self._web_session.post(
            f"http://{self._host}/api.cgi", json=payload
        ) as response:
            if response.status != 200:
                raise IHConnectionError("HTTP response status is unexpected (not 200)")

            json_response = await response.json()

        if not json_response["success"]:
            if json_response["error"]["code"] in [1, 5]:
                if self._connection_retries:
                    raise IHAuthenticationError(json_response["error"]["message"])

                # Try to reauthenticate once
                _LOGGER.debug("Request failed. Trying to reauthenticate.")
                self._connection_retries += 1
                await self._authenticate()
                return await self._request(command, **kwargs)
            raise IHConnectionError(json_response["error"]["message"])

        self._connection_retries = 0
        return json_response.get("data")

    async def _request_value(self, name: str) -> dict:
        """Get entity value by uid."""
        response = await self._request(
            LOCAL_CMD_GET_DP_VALUE, uid=COMMAND_MAP[name]["uid"]
        )
        return response["dpval"]["value"]

    async def _set_value(self, device_id, uid, value):
        return await self._request(
            LOCAL_CMD_SET_DP_VALUE,
            uid=uid,
            value=value,
        )

    async def get_datapoints(self) -> dict:
        """Get all available datapoints."""
        response = await self._request(LOCAL_CMD_GET_AVAIL_DP)
        self._datapoints = {
            dpoint["uid"]: dpoint for dpoint in response["dp"]["datapoints"]
        }
        if self._device_id:
            for dpoint in response["dp"]["datapoints"]:
                dpoint_name = INTESIS_MAP[dpoint["uid"]]["name"]
                self._devices[self._device_id][dpoint_name] = None
        return self._datapoints

    def _has_datapoint(self, datapoint: str):
        """Entity has a datapoint."""
        if datapoint not in COMMAND_MAP:
            return False
        return COMMAND_MAP[datapoint]["uid"] in self._datapoints

    async def connect(self):
        """Connect to the device and start periodic updater."""
        await self.poll_status()
        _LOGGER.debug("Successful authenticated and polled. Fetching Datapoints.")
        await self.get_datapoints()

        _LOGGER.debug("Starting updater task.")
        self._update_task = asyncio.create_task(self._run_updater())
        self._connected = True

    async def stop(self):
        """Disconnect and stop periodic updater."""
        _LOGGER.debug("Stopping updater task.")
        await self._cancel_task_if_exists(self._update_task)
        self._connected = False

        if self._own_session:
            await self._web_session.close()

    async def get_info(self) -> dict:
        """Get device info."""
        response = await self._request(LOCAL_CMD_GET_INFO)
        self._info = response["info"]
        return self._info

    async def poll_status(self, sendcallback=False):
        await self._authenticate()
        info = await self.get_info()
        self._device_id = info["sn"]
        self._controller_id = info["sn"].lower()
        self._controller_name = f"{self._info['deviceModel']} ({info['ownSSID']})"
        # Setup devices
        self._devices[self._device_id] = {
            "name": info["ownSSID"],
            "widgets": [],
            "model": info["deviceModel"],
        }

        await self.get_datapoints()
        _LOGGER.debug(repr(self._devices))

        self._update_device_state(self._device_id, "acStatus", info["acStatus"])

        if sendcallback:
            await self._send_update_callback(str(self._device_id))

    def get_mode_list(self, device_id) -> list:
        """Get possible entity modes."""
        uid = COMMAND_MAP["mode"]["uid"]
        return [
            INTESIS_MAP[uid]["values"][i]
            for i in self._datapoints[uid]["descr"]["states"]
        ]

    def _get_fan_map(self, device_id):
        fan_values = sorted(self._datapoints[4]["descr"]["states"])
        for values in INTESIS_MAP[67]["values"].values():
            if sorted(values.keys()) == fan_values:
                return values
        return INTESIS_MAP[67]["values"][63]

    def has_vertical_swing(self, device_id) -> bool:
        """Entity supports vertical swing."""
        return self._has_datapoint("vvane")

    def has_horizontal_swing(self, device_id) -> bool:
        """Entity supports horizontal swing."""
        return self._has_datapoint("hvane")

    async def _parse_response(self, decoded_data):
        """Private method to parse the API response."""
        raise NotImplementedError()



""" Constants for pyintesishome """
INTESIS_CMD_STATUS = '{"status":{"hash":"x"},"config":{"hash":"x"}}'
INTESIS_NULL = 32768

DEVICE_INTESISHOME = "IntesisHome"
DEVICE_AIRCONWITHME = "airconwithme"
DEVICE_ANYWAIR = "anywair"
DEVICE_INTESISHOME_LOCAL = "intesishome_local"
DEVICE_INTESISBOX = "IntesisBox"

API_DISCONNECTED = "Disconnected"
API_CONNECTING = "Connecting"
API_AUTHENTICATED = "Connected"
API_AUTH_FAILED = "Wrong username/password"

CONFIG_MODE_BITS = {1: "auto", 2: "heat", 4: "dry", 8: "fan", 16: "cool"}
OPERATING_MODE_BITS = {
    1: "heat",
    2: "heat+tank",
    4: "tank",
    8: "cool+tank",
    16: "cool",
    32: "auto",
    64: "auto+tank",
}

INTESIS_MAP = {
    1: {"name": "power", "values": {0: "off", 1: "on"}},
    2: {
        "name": "mode",
        "values": {0: "auto", 1: "heat", 2: "dry", 3: "fan", 4: "cool"},
    },
    4: {
        "name": "fan_speed",
    },
    5: {
        "name": "vvane",
        "values": {
            0: "auto/stop",
            1: "manual1",
            2: "manual2",
            3: "manual3",
            4: "manual4",
            5: "manual5",
            6: "manual6",
            7: "manual7",
            8: "manual8",
            9: "manual9",
            10: "swing",
        },
    },
    6: {
        "name": "hvane",
        "values": {
            0: "auto/stop",
            10: "swing",
            1: "manual1",
            2: "manual2",
            3: "manual3",
            4: "manual4",
            5: "manual5",
        },
    },
    9: {"name": "setpoint"},
    10: {"name": "temperature"},
    12: {"name": "remote_controller_lock"},
    13: {"name": "working_hours"},
    14: {"name": "alarm_status"},
    15: {"name": "error_code"},
    34: {"name": "quiet_mode", "values": {0: "off", 1: "on"}},
    35: {"name": "setpoint_min"},
    36: {"name": "setpoint_max"},
    37: {"name": "outdoor_temp"},
    38: {"name": "water_outlet_temperature"},
    39: {"name": "water_inlet_temperature"},
    42: {
        "name": "climate_working_mode",
        "values": {0: "comfort", 1: "eco", 2: "powerful"},
    },
    44: {
        "name": "tank_working_mode",
        "values": {0: "comfort", 1: "eco", 2: "powerful"},
    },
    45: {"name": "tank_water_temperature"},
    46: {"name": "solar_status"},
    48: {"name": "thermoshift_heat_eco"},
    49: {"name": "thermoshift_cool_eco"},
    50: {"name": "thermoshift_heat_powerful"},
    51: {"name": "thermoshift_cool_powerful"},
    52: {"name": "thermoshift_tank_eco"},
    53: {"name": "thermoshift_tank_powerful"},
    54: {"name": "error_reset"},
    55: {"name": "heat_thermo_shift"},
    56: {"name": "cool_water_setpoint_temperature"},
    57: {"name": "tank_setpoint_temperature"},
    58: {
        "name": "operating_mode",
        "values": {
            0: "maintenance",
            1: "heat",
            2: "heat+tank",
            3: "tank",
            4: "cool+tank",
            5: "cool",
            6: "auto",
            7: "auto+tank",
        },
    },
    60: {"name": "heat_8_10"},
    61: {"name": "config_mode_map"},
    62: {"name": "runtime_mode_restrictions"},
    63: {"name": "config_horizontal_vanes"},
    64: {"name": "config_vertical_vanes"},
    65: {"name": "config_quiet"},
    66: {"name": "config_confirm_off"},
    67: {
        "name": "config_fan_map",
        "values": {
            6: {1: "low", 2: "high"},
            7: {0: "auto", 1: "low", 2: "high"},
            14: {1: "low", 2: "medium", 3: "high"},
            15: {0: "auto", 1: "low", 2: "medium", 3: "high"},
            30: {1: "quiet", 2: "low", 3: "medium", 4: "high"},
            31: {0: "auto", 1: "quiet", 2: "low", 3: "medium", 4: "high"},
            62: {1: "quiet", 2: "low", 3: "medium", 4: "high", 5: "max"},
            63: {0: "auto", 1: "quiet", 2: "low", 3: "medium", 4: "high", 5: "max"},
            126: {
                1: "speed 1",
                2: "speed 2",
                3: "speed 3",
                4: "speed 4",
                5: "speed 5",
                6: "speed 6",
            },
            127: {
                0: "auto",
                1: "speed 1",
                2: "speed 2",
                3: "speed 3",
                4: "speed 4",
                5: "speed 5",
                6: "speed 6",
            },
        },
    },
    68: {"name": "instant_power_consumption"},
    69: {"name": "accumulated_power_consumption"},
    75: {"name": "config_operating_mode"},
    77: {"name": "config_vanes_pulse"},
    80: {"name": "aquarea_tank_consumption"},
    81: {"name": "aquarea_cool_consumption"},
    82: {"name": "aquarea_heat_consumption"},
    83: {"name": "heat_high_water_set_temperature"},
    84: {"name": "heating_off_temperature"},
    87: {"name": "heater_setpoint_temperature"},
    90: {"name": "water_target_temperature"},
    95: {
        "name": "heat_interval",
        "values": {
            1: 30,
            2: 60,
            3: 90,
            4: 120,
            5: 150,
            6: 180,
            7: 210,
            8: 240,
            9: 270,
            10: 300,
            11: 330,
            12: 360,
            13: 390,
            14: 420,
            15: 450,
            16: 480,
            17: 510,
            18: 540,
            19: 570,
            20: 600,
        },
    },
    107: {"name": "aquarea_working_hours"},
    123: {"name": "ext_thermo_control", "values": {85: "off", 170: "on"}},
    124: {"name": "tank_present", "values": {85: "off", 170: "on"}},
    125: {"name": "solar_priority", "values": {85: "off", 170: "on"}},
    134: {"name": "heat_low_outdoor_set_temperature"},
    135: {"name": "heat_high_outdoor_set_temperature"},
    136: {"name": "heat_low_water_set_temperature"},
    137: {"name": "farenheit_type"},
    140: {"name": "extremes_protection_status"},
    144: {"name": "error_code"},
    148: {"name": "extremes_protection"},
    149: {"name": "binary_input"},
    153: {"name": "config_binary_input"},
    168: {"name": "uid_binary_input_on_off"},
    169: {"name": "uid_binary_input_occupancy"},
    170: {"name": "uid_binary_input_window"},
    181: {"name": "mainenance_w_reset"},
    182: {"name": "mainenance_wo_reset"},
    183: {"name": "filter_clean"},
    184: {"name": "filter_due_hours"},
    185: {"name": "uid_185"},
    186: {"name": "uid_186"},
    191: {"name": "uid_binary_input_sleep_mode"},
    50000: {
        "name": "external_led",
        "values": {0: "off", 1: "on", 2: "blinking only on change"},
    },
    50001: {"name": "internal_led", "values": {0: "off", 1: "on"}},
    50002: {"name": "internal_temperature_offset"},
    50003: {"name": "temp_limitation", "values": {0: "off", 2: "on"}},
    50004: {"name": "cool_temperature_min"},
    50005: {"name": "cool_temperature_max"},
    50006: {"name": "heat_temperature_min"},
    50007: {"name": "heat_temperature_min"},
    60002: {"name": "rssi"},
}

COMMAND_MAP = {
    "power": {"uid": 1, "values": {"off": 0, "on": 1}},
    "mode": {"uid": 2, "values": {"auto": 0, "heat": 1, "dry": 2, "fan": 3, "cool": 4}},
    "operating_mode": {
        "uid": 58,
        "values": {
            "heat": 1,
            "heat+tank": 2,
            "tank": 3,
            "cool+tank": 4,
            "cool": 5,
            "auto": 6,
            "auto+tank": 7,
        },
    },
    "climate_working_mode": {
        "uid": 42,
        "values": {"comfort": 0, "eco": 1, "powerful": 2},
    },
    "fan_speed": {"uid": 4},
    "vvane": {
        "uid": 5,
        "values": {
            "auto/stop": 0,
            "swing": 10,
            "manual1": 1,
            "manual2": 2,
            "manual3": 3,
            "manual4": 4,
            "manual5": 5,
        },
    },
    "hvane": {
        "uid": 6,
        "values": {
            "auto/stop": 0,
            "swing": 10,
            "manual1": 1,
            "manual2": 2,
            "manual3": 3,
            "manual4": 4,
            "manual5": 5,
        },
    },
    "setpoint": {"uid": 9}
    # aquarea
    ,
    "quiet": {"uid": 34, "values": {"off": 0, "on": 1}},
    "tank": {"uid": 44, "values": {"comfort": 0, "eco": 1, "powerful": 2}},
    "reset_eror": {"uid": 54, "values": {"on": 1}},
    "tank_setpoint_temperature": {"uid": 57},
    "thermoshift_heat_eco": {"uid": 48, "min": 0, "max": 5},
    "thermoshift_cool_eco": {"uid": 49, "min": 0, "max": 5},  # 172
    "thermoshift_heat_powerful": {"uid": 50, "min": 0, "max": 5},
    "thermoshift_cool_powerful": {"uid": 51, "min": 0, "max": 5},  # 171
    "thermoshift_tank_eco": {"uid": 52, "min": 0, "max": 10},
    "thermoshift_tank_powerful": {"uid": 53, "min": 0, "max": 10},
    "heat_thermo_shift": {"uid": 55, "min": -5, "max": 5},
    "cool_water_setpoint_temperature": {"uid": 56},
    "heat_high_water_set_temperature": {"uid": 83, "min": 25, "max": 55},
    "heat_low_outdoor_set_temperature": {"uid": 134, "min": -15, "max": 15},
    "heat_high_outdoor_set_temperature": {"uid": 135, "min": -15, "max": 15},
    "heat_low_water_set_temperature": {"uid": 136, "min": 25, "max": 55},
    "resync": {"uid": 143, "values": {"on": 1}},
    "remote_control_block": {"uid": 12, "values": {"on": 2, "off": 0}},
}

# aquarea
ERROR_MAP = {
    0: {"code": "H00", "desc": "No abnormality detected"},
    2: {"code": "H91", "desc": "Tank booster heater OLP abnormality"},
    13: {"code": "F38", "desc": "Unknown"},
    20: {"code": "H90", "desc": "Indoor / outdoor abnormal communication"},
    36: {"code": "H99", "desc": "Indoor heat exchanger freeze prevention"},
    38: {"code": "H72", "desc": "Tank temperature sensor abnormality"},
    42: {"code": "H12", "desc": "Indoor / outdoor capacity unmatched"},
    156: {"code": "H76", "desc": "Indoor - control panel communication abnormality"},
    193: {"code": "F12", "desc": "Pressure switch activate"},
    195: {"code": "F14", "desc": "Outdoor compressor abnormal rotation"},
    196: {"code": "F15", "desc": "Outdoor fan motor lock abnormality"},
    197: {"code": "F16", "desc": "Total running current protection"},
    200: {"code": "F20", "desc": "Outdoor compressor overheating protection"},
    202: {"code": "F22", "desc": "IPM overheating protection"},
    203: {"code": "F23", "desc": "Outdoor DC peak detection"},
    204: {"code": "F24", "desc": "Refrigerant cycle abnormality"},
    205: {"code": "F27", "desc": "Pressure switch abnormality"},
    207: {"code": "F46", "desc": "Outdoor current transformer open circuit"},
    208: {"code": "F36", "desc": "Outdoor air temperature sensor abnormality"},
    209: {"code": "F37", "desc": "Indoor water inlet temperature sensor abnormality"},
    210: {"code": "F45", "desc": "Indoor water outlet temperature sensor abnormality"},
    212: {
        "code": "F40",
        "desc": "Outdoor discharge pipe temperature sensor abnormality",
    },
    214: {"code": "F41", "desc": "PFC control"},
    215: {
        "code": "F42",
        "desc": "Outdoor heat exchanger temperature sensor abnormality",
    },
    216: {"code": "F43", "desc": "Outdoor defrost temperature sensor abnormality"},
    222: {"code": "H95", "desc": "Indoor / outdoor wrong connection"},
    224: {"code": "H15", "desc": "Outdoor compressor temperature sensor abnormality"},
    225: {
        "code": "H23",
        "desc": "Indoor refrigerant liquid temperature sensor abnormality",
    },
    226: {"code": "H24", "desc": "Unknown"},
    227: {"code": "H38", "desc": "Indoor / outdoor mismatch"},
    228: {"code": "H61", "desc": "Unknown"},
    229: {"code": "H62", "desc": "Water flow switch abnormality"},
    230: {"code": "H63", "desc": "Refrigerant low pressure abnormality"},
    231: {"code": "H64", "desc": "Refrigerant high pressure abnormality"},
    232: {"code": "H42", "desc": "Compressor low pressure abnormality"},
    233: {"code": "H98", "desc": "Outdoor high pressure overload protection"},
    234: {"code": "F25", "desc": "Cooling / heating cycle changeover abnormality"},
    235: {"code": "F95", "desc": "Cooling high pressure overload protection"},
    236: {"code": "H70", "desc": "Indoor backup heater OLP abnormality"},
    237: {"code": "F48", "desc": "Outdoor EVA outlet temperature sensor abnormality"},
    238: {
        "code": "F49",
        "desc": "Outdoor bypass outlet temperature sensor abnormality",
    },
    65535: {"code": "N/A", "desc": "Communication error between PA-IntesisHome"},
}

API_URL = {
    DEVICE_AIRCONWITHME: "https://user.airconwithme.com/api.php/get/control",
    DEVICE_ANYWAIR: "https://anywair.intesishome.com/api.php/get/control",
    DEVICE_INTESISHOME: "https://user.intesishome.com/api.php/get/control",
}

API_VER = {
    DEVICE_AIRCONWITHME: "1.6.2",
    DEVICE_ANYWAIR: "2.9",
    DEVICE_INTESISHOME: "1.2.2",
}

LOCAL_CMD_LOGIN = "login"
LOCAL_CMD_GET_INFO = "getinfo"
LOCAL_CMD_SET_DP_VALUE = "setdatapointvalue"
LOCAL_CMD_GET_DP_VALUE = "getdatapointvalue"
LOCAL_CMD_GET_AVAIL_DP = "getavailabledatapoints"


INTESISBOX_CMD_GET_INFO = "ID"
INTESISBOX_CMD_GET_AVAIL_DP = "GET,1:*"
INTESISBOX_CMD_ONOFF = "ONOFF"
INTESISBOX_CMD_MODE = "MODE"
INTESISBOX_CMD_SETPOINT = "SETPTEMP"
INTESISBOX_CMD_FANSP = "FANSP"
INTESISBOX_CMD_VANEUD = "VANEUD"
INTESISBOX_CMD_VANELR = "VANELR"
INTESISBOX_CMD_AMBTEMP = "AMBTEMP"
INTESISBOX_CMD_ERRSTATUS = "ERRSTATUS"
INTESISBOX_CMD_ERRCODE = "ERRCODE"

INTESISBOX_CMD_MAP = {}

INTESISBOX_MODE_MAP = {
    "auto": "AUTO",
    "heat": "HEAT",
    "dry": "DRY",
    "fan": "FAN",
    "cool": "COOL",
}

INTESISBOX_MAP = {
    INTESISBOX_CMD_ONOFF: "power",
    INTESISBOX_CMD_MODE: "mode",
    INTESISBOX_CMD_SETPOINT: "setpoint",
    INTESISBOX_CMD_FANSP: "fan_speed",
    INTESISBOX_CMD_VANEUD: "vvane",
    INTESISBOX_CMD_VANELR: "hvane",
    INTESISBOX_CMD_AMBTEMP: "temperature",
    INTESISBOX_CMD_ERRSTATUS: "error_status",
    INTESISBOX_CMD_ERRCODE: "error_code",
}


INTESISBOX_INIT = [
    "ID",
    "LIMITS:SETPTEMP",
    "LIMITS:FANSP",
    "LIMITS:MODE",
    "LIMITS:VANEUD",
    "LIMITS:VANELR",
    "GET,1:ONOFF",
    "GET,1:MODE",
    "GET,1:AMBTEMP",
    "GET,1:SETPTEMP",
]



"""IntesisBox class."""
import asyncio
import logging
from typing import List

from pyintesishome.intesisbase import IntesisBase

from .const import (
    DEVICE_INTESISBOX,
    INTESIS_NULL,
    INTESISBOX_CMD_FANSP,
    INTESISBOX_CMD_GET_AVAIL_DP,
    INTESISBOX_CMD_MODE,
    INTESISBOX_CMD_ONOFF,
    INTESISBOX_CMD_SETPOINT,
    INTESISBOX_CMD_VANELR,
    INTESISBOX_CMD_VANEUD,
    INTESISBOX_INIT,
    INTESISBOX_MAP,
    INTESISBOX_MODE_MAP,
)
from .helpers import uint32

_LOGGER = logging.getLogger("pyintesishome")

# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-public-methods


class IntesisBox(IntesisBase):
    """IntesisBox local class."""

    INTESIS_MAP = INTESISBOX_MAP

    def __init__(self, host, loop=None):
        """Initialize IntesisBox controller."""
        super().__init__(host=host, loop=loop, device_type=DEVICE_INTESISBOX)
        self._scan_interval = 60
        self._device_id: str = ""
        self._values: dict = {}
        self._info: dict = {}
        self._mac: str = ""
        self._port = 3310
        self._last_command: str = ""
        self._data_delimiter = b"\r"

    async def connect(self):
        """Public method for making the controller connect."""
        if self._connected:
            _LOGGER.debug("Already connected")
            return

        self._devices = {}
        await self._cancel_task_if_exists(self._receive_task)

        _LOGGER.debug("Connecting")
        try:
            self._reader, self._writer = await asyncio.open_connection(
                self._host, self._port, loop=self._event_loop
            )
            self._receive_task = asyncio.create_task(self._data_received())
            await asyncio.wait_for(self._initialise_connection(), timeout=60.0)
            await self._send_update_callback()

        except OSError as exc:  # pylint: disable=broad-except
            _LOGGER.debug("Exception opening connection")
            _LOGGER.error("%s Exception. %s / %s", type(exc), repr(exc.args), exc)

    async def _initialise_connection(self):
        """Requests the current state of the device"""
        for init_cmd in INTESISBOX_INIT:
            await self._send_command(init_cmd)

        initialized = False

        while not initialized:
            _LOGGER.debug("Awaiting initialisation (current temperature)")
            await asyncio.sleep(0.1)
            initialized = self._device_id in self._devices and (
                "temperature" in self._devices[self._device_id]
                or "hvane_list" in self._devices[self._device_id]
            )
        self._connected = True

    async def _parse_response(self, decoded_data):
        """Parses the API response and routes to the correct handler method."""
        lines_received: List[str] = decoded_data.strip().splitlines()
        for line in lines_received:
            cmdlist = line.split(":", 1)
            cmd = cmdlist[0]
            if cmd in ["CHN,1", "ID", "LIMITS", "ACK", "ERR"]:
                if len(cmdlist) > 1:
                    args = cmdlist[1]
                    if cmd == "ID":
                        self._parse_id_received(args)
                    elif cmd == "CHN,1":
                        self._parse_change_received(args)
                    elif cmd == "LIMITS":
                        self._parse_limits_received(args)
            if not self._received_response.is_set():
                _LOGGER.debug("Resolving set_value's await")
                self._received_response.set()

    def _parse_id_received(self, args):
        # ID:Model,MAC,IP,Protocol,Version,RSSI
        info = args.split(",")
        if len(info) >= 6:
            self._info["deviceModel"] = info[0]
            self._mac = info[1]
            self._device_id = info[1]
            self._info["firmware"] = info[4]
            self._info["rssi"] = info[5]
        self._controller_id = self._mac.lower()
        self._controller_name = f"{self._info['deviceModel']} ({self._mac[-4:]})"

        # Setup devices
        if self._device_id not in self._devices:
            self._devices[self._device_id] = {
                "name": f"{self._device_type} {self._mac[-4:]}",
                "widgets": [],
                "model": self._info["deviceModel"],
            }
        _LOGGER.debug(repr(self._devices))

    def _parse_change_received(self, args):
        function, value = args.split(",")

        if value == INTESIS_NULL:
            value = None

        self._devices[self._device_id][
            INTESISBOX_MAP.get(function, function.lower())
        ] = value.lower()

    def _parse_limits_received(self, args):
        split_args = args.split(",", 1)

        if len(split_args) == 2:
            function = split_args[0]
            values = split_args[1][1:-1].lower().split(",")

            if function == INTESISBOX_CMD_SETPOINT and len(values) == 2:
                self._devices[self._device_id]["setpoint_min"] = int(values[0])
                self._devices[self._device_id]["setpoint_max"] = int(values[1])
            elif function == INTESISBOX_CMD_FANSP:
                self._devices[self._device_id]["config_fan_map"] = values
            elif function == INTESISBOX_CMD_MODE:
                self._devices[self._device_id]["mode_list"] = values
            elif function == INTESISBOX_CMD_VANEUD:
                self._devices[self._device_id]["vvane_list"] = values
            elif function == INTESISBOX_CMD_VANELR:
                self._devices[self._device_id]["hvane_list"] = values

    async def set_mode(self, device_id, mode: str):
        """Internal method for setting the mode with a string value."""
        if mode in INTESISBOX_MODE_MAP:
            await self._set_value(device_id, "MODE", INTESISBOX_MODE_MAP[mode])

    async def _request_values(self) -> dict:
        """Get all entity values."""
        raise NotImplementedError()
        # response = await self._request(LOCAL_CMD_GET_DP_VALUE, uid="all")
        # self._values = {dpval["uid"]: dpval["value"] for dpval in response["dpval"]}
        # return self._values

    async def _send_keepalive(self):
        """Run a loop that updates the values every _scan_interval."""
        try:
            while True:
                await asyncio.sleep(30)
                await self._send_command("GET,1:AMBTEMP")
        except asyncio.CancelledError:
            _LOGGER.debug("Cancelled the keepalive task")

    async def _authenticate(self) -> bool:
        """Authenticate using username and password."""
        raise NotImplementedError()

    async def _set_value(self, device_id, uid, value):
        """Internal method to send a command to the API"""
        command = f"SET,1:{uid},{value}\r"
        await self._send_command(command)

    async def set_power_off(self, device_id=None):
        """Public method to turn off the device asynchronously."""
        await self._set_value(device_id, INTESISBOX_CMD_ONOFF, "OFF")

    async def set_power_on(self, device_id=None):
        """Public method to turn on the device asynchronously."""
        await self._set_value(device_id, INTESISBOX_CMD_ONOFF, "ON")

    async def set_temperature(self, device_id, setpoint):
        """Public method for setting the temperature"""
        set_temp = uint32(setpoint * 10)
        await self._set_value(device_id, INTESISBOX_CMD_SETPOINT, set_temp)

    async def set_fan_speed(self, device_id, fan: str):
        """Public method to set the fan speed"""
        await self._set_value(device_id, INTESISBOX_CMD_FANSP, fan)

    async def set_vertical_vane(self, device_id, vane: str):
        """Public method to set the vertical vane"""
        await self._set_value(device_id, INTESISBOX_CMD_VANEUD, vane)

    async def set_horizontal_vane(self, device_id, vane: str):
        """Public method to set the horizontal vane"""
        await self._set_value(device_id, INTESISBOX_CMD_VANELR, vane)

    async def poll_status(self, sendcallback=False):
        if self._connected:
            _LOGGER.debug("Polling status")
            await self._send_command(INTESISBOX_CMD_GET_AVAIL_DP)

    def get_mode_list(self, device_id) -> list:
        """Get possible entity modes."""
        return self._devices[self._device_id].get("mode_list")

    def _get_fan_map(self, device_id):
        fan_map_list = self._devices[self._device_id].get("config_fan_map")
        if isinstance(fan_map_list, list):
            return dict(zip(fan_map_list, fan_map_list))
        return None



#!/usr/bin/env python3
from setuptools import setup

setup(
    name="pyintesishome",
    version="1.8.1",
    description="A python3 library for running asynchronus communications with IntesisHome Smart AC Controllers",
    url="https://github.com/jnimmo/pyIntesisHome",
    author="James Nimmo",
    author_email="james@nimmo.net.nz",
    license="MIT",
    install_requires=["aiohttp>=3.7.4,<4"],
    packages=["pyintesishome"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    ],
)
